from typing import Tuple

from sparql_slurper import SlurpyGraph, TM_NS

graphdb_id = "<http://www.ontotext.com/owlim/entity#id>"


class GraphDBSlurpyGraph(SlurpyGraph):

    def gen_query(self,  pattern, gquery: str, gqueryend: str) -> str:
        """ Generate a query that includes the identifiers of any variables and adds identifier translation for
        any URI's
        """
        def gen_s_o(p: str, is_s: bool) -> Tuple[str, str]:
            tgt = '?s' if is_s else '?o'
            if p is None:
                return tgt, f' {tgt} {graphdb_id} {tgt}id . '
            elif str(p).startswith(str(TM_NS)):
                bnode_id = str(p)[len(str(TM_NS)):]
                return tgt, f' {tgt} {graphdb_id} {bnode_id} . '
            else:
                return self._repr_element(p), ''

        subj, s_add = gen_s_o(pattern[0], True)
        pred = self._repr_element(pattern[1]) if pattern[1] is not None else '?p'
        obj, o_add = gen_s_o(pattern[2], False)
        return f"SELECT ?s ?p ?o ?sid ?oid {{{gquery}{subj} {pred} {obj} . {s_add}{o_add} {gqueryend}}}"



from pbr.version import VersionInfo

# A hand agent to use for testing
__version__ = VersionInfo('sparql_slurper')
UserAgent = f"{__version__.package}/{__version__.version_string()} " \
            f"(https://github.com/hsolbrig/PyShEx; solbrig@jhu.edu)"


import time
from typing import Dict, NamedTuple, Union, List, Tuple, Optional, Type
from urllib.parse import urlsplit, parse_qsl, urlunsplit

from SPARQLWrapper import SPARQLWrapper, JSON
from rdflib import Graph, URIRef, Literal, BNode, Namespace

QueryTriple = Tuple[Optional[URIRef], Optional[URIRef], Optional[Union[Literal, URIRef]]]


NodeType = Union[URIRef, Literal, BNode]


class RDFTriple(NamedTuple):
    s: Union[URIRef, BNode]
    p: URIRef
    o: Union[Literal, URIRef, BNode]


TM_NS = Namespace("http://www.ontotext.com/tm#")


class SlurpyGraph(Graph):
    """ A Graph that acts as a "cache" for a SPARQL endpoint """
    def __init__(self, endpoint: str, *args, persistent_bnodes: bool = False, agent: Optional[str] = None,
                 **kwargs) -> None:
        """ Create a graph

        :param endpoint: URL of SPARQL endpoint
        :param persistent_bnodes: BNodes persist across SPARQL calls,
        :param agent: User agent
        """
        endpoint_base, query = self._parse_endpoint_parms(endpoint)
        self.sparql = SPARQLWrapper(endpoint_base)
        for k, v in query:
            self.sparql.addParameter(k, v)
        self.persistent_bnodes = persistent_bnodes
        self.sparql.setReturnFormat(JSON)
        self.resolved_nodes: List[QueryTriple] = [(None, None, None)]
        self.debug_slurps = False
        self._query_result_hook: Optional[Type[QueryResultHook]] = None
        self.graph_name: Optional[str] = None
        self.total_slurptime = 0.0
        self.total_calls = 0
        self.total_queries = 0
        self.total_triples = 0
        self.sparql_locked = False
        super().__init__(*args, **kwargs)
        if agent:
            self.sparql.agent = agent

    def _parse_endpoint_parms(self, endpoint: str) -> Tuple[str, List[Tuple[str, str]]]:
        """
        Split the parameters off of endpoint and pass them to SPARQLWrapper via the addParams option
        :param endpoint: Endpoint URI including possible parameter strings
        :return: Endpoint URI and separate args
        """
        x = urlsplit(endpoint, allow_fragments=False)
        query = parse_qsl(x.query)
        return urlunsplit((x.scheme, x.netloc, x.path, '', x.fragment)), query

    def add_result_hook(self, hook: Type["QueryResultHook"]) -> Type["QueryResultHook"]:
        """
        Add a query result hook to the chain
        :param hook: hook to add
        :return: added hook (same as hook to add)
        """
        hook.next_hook = self._query_result_hook
        self._query_result_hook = hook
        return hook

    def _map_type(self, node: Dict, node_id: Optional[Dict] = None) -> NodeType:
        """
         Return an rdflib representation of node
        :param node: node value
        :param node_id: permanant identifier of node if applicable
        :return: rdflib type
        """
        if not self.persistent_bnodes and node['type'] == 'bnode' and node_id is None:
            raise ValueError("SlurpyGraph cannot process BNodes")
        return URIRef(node['value']) if node['type'] == 'uri' else \
            (BNode(node['value']) if node_id is None else TM_NS[node_id['value']]) if node['type'] == 'bnode' else \
            Literal(node['value'], datatype=node.get('datatype'))

    @staticmethod
    def _repr_element(node: Union[URIRef, BNode, Literal]) -> str:
        return f"<{node}>" if isinstance(node, URIRef) else f"_:{node}" if isinstance(node, BNode) else '"{node}"'

    def already_resolved(self, pattern: QueryTriple) -> bool:
        """ Determine whether pattern has already been loaded into the cache.

        The "wild card" - `(None, None, None)` - always counts as resolved.

        :param pattern: pattern to check
        :return: True it is a subset of elements already loaded
        """
        if self.sparql_locked or pattern == (None, None, None):
            return True
        for resolved_node in self.resolved_nodes:
            if resolved_node != (None, None, None) and \
                    (pattern[0] == resolved_node[0] or resolved_node[0] is None) and \
                    (pattern[1] == resolved_node[1] or resolved_node[1] is None) and\
                    (pattern[2] == resolved_node[2] or resolved_node[2] is None):
                return True
        return False

    def gen_query(self, pattern, gquery: str, gqueryend: str) -> str:
        subj = self._repr_element(pattern[0]) if pattern[0] is not None else '?s'
        pred = self._repr_element(pattern[1]) if pattern[1] is not None else '?p'
        obj = self._repr_element(pattern[2]) if pattern[2] is not None else '?o'
        return f"SELECT ?s ?p ?o {{{gquery}{subj} {pred} {obj}{gqueryend}}}"

    def triples(self, pattern: QueryTriple):
        """ Return the triples that match pattern

        :param pattern: `(s, p, o)` tuple, with `None` as wild cards
        :return: Generator for resulting triples
        """
        self.total_calls += 1
        if self.graph_name is not None:
            gn = "?g" if not self.graph_name else self.graph_name
            gquery = f"graph {gn} {{"
            gqueryend = '}'
        else:
            gquery = gqueryend = ''
        if not self.already_resolved(pattern):
            query = self.gen_query(pattern, gquery, gqueryend)
            start = time.time()
            if self.debug_slurps:
                print(f"SPARQL: ({query})", end="")
            self.sparql.setQuery(query)
            resp = self.sparql.query().convert()
            elapsed = time.time() - start
            ntriples = len(resp['results']['bindings'])
            self.total_slurptime += elapsed
            self.total_triples += ntriples
            self.total_queries += 1
            if self.debug_slurps:
                print(f" ({round(elapsed, 2)} secs) - {ntriples} triples")
            query_result = self._query_result_hook(self) if self._query_result_hook is not None else None
            for row in resp['results']['bindings']:
                triple = RDFTriple(pattern[0] if pattern[0] is not None else self._map_type(row['s'], row.get('sid', None)),
                                   pattern[1] if pattern[1] is not None else self._map_type(row['p']),
                                   pattern[2] if pattern[2] is not None else self._map_type(row['o'], row.get('oid', None)))
                self.add(triple)
                if query_result:
                    query_result.add(triple)
            if query_result:
                query_result.done()
            self.resolved_nodes.append(pattern)
        return super().triples(pattern)

    def serialize(self, destination=None, format="xml",
                  base=None, encoding=None, **args):
        self.sparql_locked = True
        try:
            rval = super().serialize(destination, format, base, encoding, **args)
        finally:
            self.sparql_locked = False
        return rval


class QueryResultHook:
    """
    QueryResultHook: used for printing and other processing out the outcome of a slurper query
    """
    next_hook = None

    def __init__(self, sg: SlurpyGraph) -> None:
        self.chained_hook = self.next_hook(sg) if self.next_hook is not None else None

    def add(self, t: RDFTriple) -> None:
        """
        Add a triple as a query result
        :param t: triple being added
        """
        if self.chained_hook is not None:
            self.chained_hook.add(t)

    def done(self) -> None:
        """
        All triples were added for this result
        """
        if self.chained_hook is not None:
            self.chained_hook.done()


class QueryResultPrinter(QueryResultHook):
    """
    Query result hook to print query results
    """
    include_namespaces = False

    def __init__(self, sg: SlurpyGraph) -> None:
        self.g = Graph(namespace_manager=sg.namespace_manager)
        super().__init__(sg)

    def add(self, t: RDFTriple) -> None:
        self.g.add(t)
        super().add(t)

    def done(self) -> None:
        g_str = self.g.serialize(format="turtle").decode()
        print('RESULTS:\n\t' + '\n\t'.join([l for l in g_str.split('\n')
                                            if self.include_namespaces or (l and not l.startswith('@prefix'))]))



import unittest

from rdflib import Namespace, BNode

from sparql_slurper import SlurpyGraph, URIRef, Literal

NEXTPROT = Namespace("http://nextprot.org/rdf#")


class PersistentBnodeTestCase(unittest.TestCase):
    def test_persistent_bnode(self):
        """ Test the persistent_bnode argument """
        g = SlurpyGraph("http://sparql-playground.nextprot.org/sparql")
        with self.assertRaises(ValueError):
            _ = list(g.objects(NEXTPROT['NATUR-NX_PEPT00119967-NX_Q9Y5X1-1'], NEXTPROT.evidence))
        g.persistent_bnodes = True
        evidence = g.value(NEXTPROT['NATUR-NX_PEPT00119967-NX_Q9Y5X1-1'], NEXTPROT.evidence)
        self.assertTrue(isinstance(evidence, BNode))
        peptide_name = g.value(evidence, NEXTPROT.peptideName)
        if peptide_name != "NX_PEPT00119967":
            print(f"Warning: {g.sparql.endpoint} does not support persistent BNodes!")


if __name__ == '__main__':
    unittest.main()



import unittest

import io
from contextlib import redirect_stdout

from rdflib import Namespace, Graph
from sparql_slurper import SlurpyGraph
from sparql_slurper._graphdb_slurpygraph import GraphDBSlurpyGraph

endpoint = 'https://graph.fhircat.org/repositories/fhirontology'


class GraphDBBnodeTestCase(unittest.TestCase):

    def test_bnodes(self):
        FHIR = Namespace("http://hl7.org/fhir/")
        P = Namespace("http://hl7.org/fhir/Patient/")
        g = GraphDBSlurpyGraph(endpoint)
        g.debug_slurps = False
        obj = g.value(P.pat4, FHIR['Resource.id'])
        v = g.value(obj, FHIR.value)
        self.assertEqual(str(v), "pat4")


if __name__ == '__main__':
    unittest.main()



import io
import unittest
from contextlib import redirect_stdout
from typing import List

from rdflib import Namespace

from sparql_slurper import SlurpyGraph, QueryResultPrinter, QueryResultHook, RDFTriple
from tests import UserAgent

endpoint = 'https://query.wikidata.org/sparql'
WD = Namespace("http://www.wikidata.org/entity/")


class PrintOptionsTestCase(unittest.TestCase):
    @staticmethod
    def setup_test() -> SlurpyGraph:
        g = SlurpyGraph(endpoint, agent=UserAgent)
        g.bind("wd", WD)
        g.bind("wdp", "http://www.wikidata.org/prop/")
        g.bind("wdd", "http://www.wikidata.org/prop/direct/")
        g.bind("sdo", "http://schema.org/")
        g.bind("wds", "http://www.wikidata.org/entity/statement/")
        g.bind("wikiba", "http://wikiba.se/ontology#")
        g.add_result_hook(QueryResultPrinter)
        return g

    @staticmethod
    def do_test(g: SlurpyGraph) -> str:
        output = io.StringIO()
        with redirect_stdout(output):
            _ = list(g.predicate_objects(WD.Q29017194))
        return output.getvalue()

    def valid_print_options_output(self, result: str) -> None:
        if not result.startswith("RESULTS:\n\twd:Q29017194 "):
            print(result)
            self.assertTrue(False, "Unexpected results")

    def test_print_options(self):
        """ Test the basic print_option option """

        # the contents of the graph can change, so just make sure we're printing something
        QueryResultPrinter.include_namespaces = False
        g = self.setup_test()
        self.valid_print_options_output(self.do_test(g))

    def test_print_with_namespace(self):
        """ Test print including namespaces and chain function """
        QueryResultPrinter.include_namespaces = True
        g = self.setup_test()
        result = self.do_test(g)
        if not result.startswith('RESULTS:\n\t@prefix rdf: ') or '\twd:Q29017194 ' not in result:
            print(result)
            self.assertTrue(False, "Unexpected results")

    def test_hook_chaining(self):
        """ Test the hook chain function """
        nadded = [0]

        class ChainedHook(QueryResultHook):
            def __init__(self, sg: SlurpyGraph):
                super().__init__(sg)

            def add(self, t: RDFTriple) -> None:
                assert isinstance(t, RDFTriple)
                nadded[0] = nadded[0] + 1
                super().add(t)

            def done(self) -> None:
                super().done()
                nadded.append(True)

        QueryResultPrinter.include_namespaces = False
        g = self.setup_test()
        g.add_result_hook(ChainedHook)
        result = self.do_test(g)
        self.valid_print_options_output(result)
        self.assertTrue(nadded[0] > 10)
        self.assertTrue(nadded[1])


if __name__ == '__main__':
    unittest.main()



import unittest

import io
from contextlib import redirect_stdout

from rdflib import Namespace, Graph
from sparql_slurper import SlurpyGraph
from tests import UserAgent

endpoint = 'https://query.wikidata.org/sparql'


class ReactomeTestCase(unittest.TestCase):

    def test_reactome(self):
        WD = Namespace("http://www.wikidata.org/entity/")
        P = Namespace("http://www.wikidata.org/prop/")
        g = SlurpyGraph(endpoint, agent=UserAgent)
        g.debug_slurps = False
        gpo = list(g.predicate_objects(WD.Q29017194))
        orig_len = len(gpo)
        # Test that we get something
        self.assertTrue(30 < orig_len < 60)                 # Current value is 33
        # Test that it is reproducable
        self.assertEqual(orig_len, len(list(g.predicate_objects((WD.Q29017194)))))
        # Test that it is cached
        self.assertEqual(orig_len, len(g))
        # Add a path
        added_length = 0
        for s in g.objects(WD.Q29017194, P.P31):
            gp31s = list(g.predicate_objects(s))
            added_length += len(gp31s)
        self.assertEqual(orig_len + added_length, len(g))
        print(f"{g.total_calls}, {g.total_queries}, {g.total_slurptime}, {g.total_triples}")
        self.assertEqual((5, 3, 41), (g.total_calls, g.total_queries, g.total_triples))
        self.assertTrue(g.total_slurptime > 0.1)

    def test_debug_slurps(self):
        WD = Namespace("http://www.wikidata.org/entity/")
        g = SlurpyGraph(endpoint, agent=UserAgent)
        g.debug_slurps = True
        output = io.StringIO()
        with redirect_stdout(output):
            _ = list(g.predicate_objects(WD.Q29017194))
        if not output.getvalue().startswith("SPARQL: "
                                            "(SELECT ?s ?p ?o {<http://www.wikidata.org/entity/Q29017194> ?p ?o})"):
            print("Unexpected:")
            print(output.getvalue())
            self.assertTrue(False)

    def test_serialize(self):
        # Issue #1 - serialize goes after the entire graph
        WD = Namespace("http://www.wikidata.org/entity/")
        g = SlurpyGraph(endpoint, agent=UserAgent)
        _ = list(g.predicate_objects(WD.Q29017194))
        g.debug_slurps = True
        # Warning - this will go away forever if the bug isn't fixed
        serialized_graph = g.serialize(format="turtle").decode('utf-8')
        g2 = Graph()
        g2.parse(data=serialized_graph, format="turtle")
        self.assertEqual(len(list(g)), len(list(g2)))



if __name__ == '__main__':
    unittest.main()



import unittest

from rdflib import Namespace, Literal

from sparql_slurper import SlurpyGraph


class AlreadyResolvedTestCase(unittest.TestCase):
    def test_already_resolved(self):
        g = SlurpyGraph("endpoint")
        EX = Namespace("http://example.org/")
        g.resolved_nodes.append((EX.s1, EX.p1, EX.o1))
        g.resolved_nodes.append((EX.s1, EX.p1, Literal("Steam")))
        g.resolved_nodes.append((EX.s1, EX.p1, Literal(117)))
        g.resolved_nodes.append((EX.s2, EX.p2, None))
        g.resolved_nodes.append((EX.s3, None, EX.o3))
        g.resolved_nodes.append((EX.s3, None, Literal("Gas")))
        g.resolved_nodes.append((EX.s3, None, Literal(-42)))
        g.resolved_nodes.append((EX.s4, None, None))
        g.resolved_nodes.append((None, EX.p5, EX.o5))
        g.resolved_nodes.append((None, EX.p5, Literal("Coal")))
        g.resolved_nodes.append((None, EX.p5, Literal(11.2)))
        g.resolved_nodes.append((None, EX.p6, None))
        g.resolved_nodes.append((None, None, EX.o7))
        g.resolved_nodes.append((None, None, Literal("Wood")))
        g.resolved_nodes.append((None, None, Literal(3.14159)))
        self.assertTrue(g.already_resolved((None, None, None)))

        self.assertTrue(g.already_resolved((EX.s1, EX.p1, EX.o1)))
        self.assertFalse(g.already_resolved((EX.s1, EX.p1, EX.o2)))
        self.assertTrue(g.already_resolved((EX.s1, EX.p1, Literal("Steam"))))
        self.assertFalse(g.already_resolved((EX.s1, EX.p1, Literal("Gas"))))
        self.assertTrue(g.already_resolved((EX.s1, EX.p1, Literal(117))))
        self.assertFalse(g.already_resolved((EX.s1, EX.p1, Literal(118))))
        self.assertFalse(g.already_resolved((EX.s1, None, Literal(117))))
        self.assertFalse(g.already_resolved((None, EX.p1, Literal(117))))

        self.assertTrue(g.already_resolved((EX.s2, EX.p2, Literal(143))))
        self.assertFalse(g.already_resolved((EX.s2, None, Literal(143))))
        self.assertTrue(g.already_resolved((EX.s4, EX.p4, None)))
        self.assertTrue(g.already_resolved((EX.s4, None, Literal(12))))
        self.assertTrue(g.already_resolved((EX.s4, None, None)))

        self.assertTrue(g.already_resolved((EX.s5, EX.p5, EX.o5)))
        self.assertFalse(g.already_resolved((EX.s5, EX.p5, EX.o5n)))

        self.assertTrue(g.already_resolved((EX.s7, None, EX.o7)))
        self.assertTrue(g.already_resolved((None, EX.p7, EX.o7)))
        self.assertFalse(g.already_resolved((EX.s7, None, None)))


if __name__ == '__main__':
    unittest.main()
